/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"
#include <wordsplit.h>

static FILE *input;
const char *input_name;
unsigned input_line;
char input_buf[1024];
int input_ready;

static int
input_read_line(void)
{
	if (input_ready)
		input_ready = 0;
	else if (fgets(input_buf, sizeof(input_buf), input)) {
		size_t len = strlen(input_buf);
		if (input_buf[len-1] == '\n')
			input_buf[len-1] = 0;
		input_line++;
	} else
		return 1;
	return 0;
}
		
static char *
input_read_id(void)
{
	if (input_read_line())
		return NULL;
	if (!input_buf[0])
		return NULL;
	return input_buf;
}

static void
input_error_recovery(void)
{
	while (input_read_line() == 0)
		if (input_buf[0] == 0)
			break;
}

static void
input_read_server(struct slb_server *srv)
{
	struct wordsplit ws;
	int wsflags = WRDSF_NOVAR | WRDSF_NOCMD | WRDSF_QUOTE |
		      WRDSF_SQUEEZE_DELIMS;

	if (!srv->test_pdu)
		srv->test_pdu = snmp_pdu_create(SNMP_MSG_GET);
	while (1) {
		size_t len;
		char *p, *q;
		oid oid[MAX_OID_LEN];
		size_t oidlen;
		
		if (input_read_line())
			break;
		if (!input_buf[0])
			break;

		if (wordsplit(input_buf, &ws, wsflags)) {
			logmsg(LOG_ERR,
			       "%s:%d: cannot split line: %s",
			       input_name, input_line,
			       wordsplit_strerror(&ws));
			continue;
		}
		ws.ws_flags |= WRDSF_REUSE;

		if (ws.ws_wordc != 3) {
			logmsg(LOG_ERR,
			       "%s:%d: malformed line",
			       input_name, input_line);
			continue;
		}
		
		oidlen = MAX_OID_LEN;
		if (!read_objid(ws.ws_wordv[0], oid, &oidlen)) {
			logmsg(LOG_ERR,
			       "%s:%d: cannot parse oid",
			       input_name, input_line);
			continue;
		}

		if (snmp_add_var(srv->test_pdu, oid, oidlen,
				 *ws.ws_wordv[1], ws.ws_wordv[2])) {
			logmsg(LOG_CRIT,
			       "%s:%d: cannot add variable: %s",
			       input_name, input_line,
			       snmp_api_errstring(snmp_errno));
			exit(EX_DATAERR);
		}
	}
	wordsplit_free(&ws);
}

static int
test_server_replies()
{
	struct slb_server *srv;
	
	for (srv = srv_head; srv; srv = srv->next) {
		if (srv->flags & SLB_SRV_DISABLED)
			continue;
		
		varinst_clear(srv->varinst);
		srv->flags &= ~SLB_SRV_COMPUTED;
		
		server_process_pdu(srv->test_pdu ?
				   NETSNMP_CALLBACK_OP_RECEIVED_MESSAGE :
				   NETSNMP_CALLBACK_OP_TIMED_OUT,
				   srv,
				   srv->test_pdu);
		if (srv->test_pdu) {
			snmp_free_pdu(srv->test_pdu);
			srv->test_pdu = NULL;
		}
	}
	sort();
	slb_fmtout();
	slb_loop_serial++;
	slb_loop_ts += wakeup_interval;
}

int
slb_test(const char *name)
{
	if (!name) {
		input = stdin;
		input_name = "<stdin>";
	} else {
		input = fopen(name, "r");
		if (!input) {
			logmsg(LOG_CRIT, "cannot open input file %s: %s",
			       name, strerror(errno));
			return EX_NOINPUT;
		}
		input_name = name;
	}
	input_line = 0;
	
	while (!feof(input)) {
		const char *id;

		id = input_read_id();
		if (id) {
			struct slb_server *srv = server_lookup(id);
			if (!srv) {
				logmsg(LOG_ERR,
				       "%s:%d: unknown server id",
				       input_name, input_line);
				input_error_recovery();
				continue;
			} else if (srv->flags & SLB_SRV_DISABLED) {
				logmsg(LOG_INFO,
				       "%s:%d: inoring disabled server %s",
				       input_name, input_line, srvid(srv));
				input_error_recovery();
				continue;
			} else {
				input_read_server(srv);
			}
		} else 
			test_server_replies();
	}
	
	fclose(input);
	return EX_OK;
}
